import configparser
import datetime
import hashlib
import itertools
import json
import logging
import pathlib
import time
import urllib.parse
from typing import Any
from typing import Dict
from typing import Iterable
from typing import Iterator
from typing import List
from typing import Optional
from typing import Tuple

import msgpack
import requests
import sqlalchemy as sa
import tqdm
from typer import Context as TyperContext
from typer import Option
from typer import Exit
from typer import echo

from spinta import exceptions
from spinta import spyna
from spinta.cli.helpers.auth import require_auth
from spinta.cli.helpers.data import ModelRow
from spinta.cli.helpers.data import count_rows
from spinta.cli.helpers.data import iter_model_rows
from spinta.cli.helpers.store import prepare_manifest
from spinta.components import Action
from spinta.components import Context
from spinta.components import Model
from spinta.types.namespace import sort_models_by_refs
from spinta.utils.data import take
from spinta.utils.json import fix_data_for_json
from spinta.utils.nestedstruct import flatten
from spinta.utils.units import tobytes
from spinta.utils.units import toseconds

log = logging.getLogger(__name__)


def push(
    ctx: TyperContext,
    target: str,
    dataset: str = Option(None, '-d', '--dataset', help=(
        "Push only specified dataset"
    )),
    credentials: str = Option(None, '-r', '--credentials', help=(
        "Credentials file"
    )),
    client: str = Option(None, '-c', '--client', help=(
        "Client name from credentials file"
    )),
    auth: str = Option(None, '-a', '--auth', help="Authorize as a client"),
    limit: int = Option(None, help=(
        "Limit number of rows read from each model"
    )),
    chunk_size: str = Option('1m', help=(
        "Push data in chunks (1b, 1k, 2m, ...), default: 1m"
    )),
    stop_time: str = Option(None, help=(
        "Stop pushing after given time (1s, 1m, 2h, ...)"
    )),
    stop_row: int = Option(None, help="Stop after pushing n rows."),
    state: pathlib.Path = Option(None, help="Save push state into a file."),
):
    """Push data to external data store"""
    if chunk_size:
        chunk_size = tobytes(chunk_size)

    if stop_time:
        stop_time = toseconds(stop_time)

    context = ctx.obj
    store = prepare_manifest(context)

    if credentials:
        credentials = pathlib.Path(credentials)
        if not credentials.exists():
            echo(f"Credentials file {credentials} does not exit.")
            raise Exit(code=1)

    manifest = store.manifest
    if dataset and dataset not in manifest.datasets:
        echo(str(exceptions.NodeNotFound(manifest, type='dataset', name=dataset)))
        raise Exit(code=1)

    ns = manifest.namespaces['']

    with context:
        require_auth(context, auth)
        context.attach('transaction', manifest.backend.transaction)

        backends = set()
        for backend in store.backends.values():
            backends.add(backend.name)
            context.attach(f'transaction.{backend.name}', backend.begin)
        for backend in manifest.backends.values():
            backends.add(backend.name)
            context.attach(f'transaction.{backend.name}', backend.begin)
        for dataset_ in manifest.datasets.values():
            for resource in dataset_.resources.values():
                if resource.backend and resource.backend.name not in backends:
                    backends.add(resource.backend.name)
                    context.attach(f'transaction.{resource.backend.name}', resource.backend.begin)
        for keymap in store.keymaps.values():
            context.attach(f'keymap.{keymap.name}', lambda: keymap)

        from spinta.types.namespace import traverse_ns_models

        models = traverse_ns_models(context, ns, Action.SEARCH, dataset)
        models = sort_models_by_refs(models)
        models = list(reversed(list(models)))
        counts = count_rows(context, models, limit)

        if state:
            engine, metadata = _init_push_state(state, models)
            context.attach('push.state.conn', engine.begin)

        rows = iter_model_rows(context, models, counts, limit)
        rows = _prepare_rows_for_push(rows)

        rows = tqdm.tqdm(rows, 'PUSH', ascii=True, total=sum(counts.values()))

        if stop_time:
            rows = _add_stop_time(rows, stop_time)

        if state:
            rows = _check_push_state(context, rows, metadata)

        if stop_row:
            rows = itertools.islice(rows, stop_row)

        rows = _push_to_remote(rows, target, credentials, client, chunk_size)

        if state:
            rows = _save_push_state(context, rows, metadata)

        while True:
            try:
                next(rows)
            except StopIteration:
                break
            except Exception:
                log.exception("Error while reading data.")


class _PushRow:
    model: Model
    data: dict
    rev: Optional[str]
    saved: bool = False

    def __init__(self, model: Model, data: Dict[str, Any]):
        self.model = model
        self.data = data
        self.rev = None
        self.saved = False


def _prepare_rows_for_push(rows: Iterable[ModelRow]) -> Iterator[_PushRow]:
    for model, row in rows:
        _id = row['_id']
        _type = row['_type']
        where = {
            'name': 'eq',
            'args': [
                {'name': 'bind', 'args': ['_id']},
                _id,
            ]
        }
        payload = {
            '_op': 'upsert',
            '_type': _type,
            '_id': _id,
            '_where': spyna.unparse(where),
            **{k: v for k, v in row.items() if not k.startswith('_')}
        }
        yield _PushRow(model, payload)


def _push_to_remote(
    rows: Iterable[_PushRow],
    target: str,
    credentials: pathlib.Path,
    client: str,
    chunk_size: int,
):
    echo(f"Get access token from {target}")
    token = _get_access_token(credentials, client, target)

    session = requests.Session()
    session.headers['Content-Type'] = 'application/json'
    session.headers['Authorization'] = f'Bearer {token}'

    prefix = '{"_data":['
    suffix = ']}'
    slen = len(suffix)
    chunk = prefix
    ready = []

    for row in rows:
        data = fix_data_for_json(row.data)
        data = json.dumps(data, ensure_ascii=False)
        if ready and len(chunk) + len(data) + slen > chunk_size:
            yield from _send_and_receive(session, target, ready, chunk + suffix)
            chunk = prefix
            ready = []
        chunk += (',' if ready else '') + data
        ready.append(row)

    if ready:
        yield from _send_and_receive(session, target, ready, chunk + suffix)


def _send_and_receive(session, target, rows: List[_PushRow], data: str):
    data = data.encode('utf-8')

    try:
        resp = session.post(target, data=data)
        resp.raise_for_status()
        data = resp.json()['_data']
    except Exception:
        log.exception("Error when sending and receiving data.")
        return

    assert len(rows) == len(data), (
        f"len(sent) = {len(rows)}, len(received) = {len(data)}"
    )
    for sent, recv in zip(rows, data):
        assert sent.data['_id'] == recv['_id'], (
            f"sent._id = {sent.data['_id']}, received._id = {recv['_id']}"
        )
        yield sent


def _add_stop_time(rows, stop):
    start = time.time()
    for row in rows:
        yield row
        if time.time() - start > stop:
            break


def _init_push_state(
    file: pathlib.Path,
    models: List[Model],
) -> Tuple[sa.engine.Engine, sa.MetaData]:
    engine = sa.create_engine(f'sqlite:///{file}')
    metadata = sa.MetaData(engine)
    for model in models:
        table = sa.Table(
            model.name, metadata,
            sa.Column('id', sa.Unicode, primary_key=True),
            sa.Column('rev', sa.Unicode),
            sa.Column('pushed', sa.DateTime),
        )
        table.create(checkfirst=True)
    return engine, metadata


def _get_model_type(row: _PushRow) -> str:
    return row.data['_type']


def _check_push_state(
    context: Context,
    rows: Iterable[_PushRow],
    metadata: sa.MetaData,
):
    conn = context.get('push.state.conn')

    for model_type, group in itertools.groupby(rows, key=_get_model_type):
        table = metadata.tables[model_type]

        query = sa.select([table.c.id, table.c.rev])
        saved = {
            state[table.c.id]: state[table.c.rev]
            for state in conn.execute(query)
        }

        for row in group:
            _id = row.data['_id']

            rev = fix_data_for_json(take(row.data))
            rev = flatten([rev])
            rev = [[k, v] for x in rev for k, v in sorted(x.items())]
            rev = msgpack.dumps(rev, strict_types=True)
            rev = hashlib.sha1(rev).hexdigest()

            row.rev = rev
            row.saved = _id in saved

            if saved.get(_id) == row.rev:
                continue  # Nothing has changed.

            yield row


def _save_push_state(
    context: Context,
    rows: Iterable[_PushRow],
    metadata: sa.MetaData,
):
    conn = context.get('push.state.conn')
    for row in rows:
        table = metadata.tables[row.data['_type']]
        if row.saved:
            conn.execute(
                table.update().
                where(table.c.id == row.data['_id']).
                values(
                    id=row.data['_id'],
                    rev=row.rev,
                    pushed=datetime.datetime.now(),
                )
            )
        else:
            conn.execute(
                table.insert().
                values(
                    id=row.data['_id'],
                    rev=row.rev,
                    pushed=datetime.datetime.now(),
                )
            )
        yield row


def _get_access_token(credsfile: pathlib.Path, client, url) -> str:
    url = urllib.parse.urlparse(url)
    section = f'{client}@{url.hostname}'
    if url.port:
        section += f':{url.port}'
    creds = configparser.ConfigParser()
    creds.read(credsfile)
    auth = (
        creds.get(section, 'client_id'),
        creds.get(section, 'client_secret'),
    )
    resp = requests.post(f'{url.scheme}://{url.netloc}/auth/token', auth=auth, data={
        'grant_type': 'client_credentials',
        'scope': creds.get(section, 'scopes'),
    })
    if resp.status_code >= 400:
        echo(resp.text)
        echo("Can't get access token.")
        raise Exit(code=1)
    return resp.json()['access_token']



import pathlib

from spinta import commands
from spinta.components import Context
from spinta.exceptions import ManifestFileDoesNotExist
from spinta.manifests.tabular.components import TabularManifest


@commands.configure.register(Context, TabularManifest)
def configure(context: Context, manifest: TabularManifest):
    rc = context.get('rc')
    path = rc.get('manifests', manifest.name, 'path')
    if path:
        path = pathlib.Path(path)
        if not path.exists():
            raise ManifestFileDoesNotExist(manifest, path=path)
    else:
        path = None
    manifest.path = path



import json
import pathlib

from ruamel.yaml import YAML

from spinta.utils.config import asbool
from spinta.utils.imports import importstr
from spinta.commands import load, check
from spinta.components import Context, Config
from spinta import components
from spinta.core.ufuncs import ufunc

yaml = YAML(typ='safe')


@load.register(Context, Config)
def load(context: Context, config: Config) -> Config:
    # Save reference to raw config and give chance for other components to read
    # directly from raw config.
    rc = config.rc = context.get('rc')

    # Load commands.
    config.commands = {}
    for scope in rc.keys('commands'):
        if scope == 'modules':
            continue
        config.commands[scope] = {}
        for name in rc.keys('commands', scope):
            command = rc.get('commands', scope, name, cast=importstr)
            config.commands[scope][name] = command

    # Load ufuncs.
    ufunc.resolver.collect(rc.get('ufuncs'))
    config.resolvers = ufunc.resolver.ufuncs()
    config.executors = ufunc.executor.ufuncs()

    # Load components.
    config.components = {}
    for group in rc.keys('components'):
        config.components[group] = {}
        for name in rc.keys('components', group):
            component = rc.get('components', group, name, cast=importstr)
            config.components[group][name] = component

    # Load exporters.
    config.exporters = {}
    for name in rc.keys('exporters'):
        exporter = rc.get('exporters', name, cast=importstr)
        config.exporters[name] = exporter()

    # Load accesslog
    accesslog = rc.get('accesslog', 'type', required=True)
    config.AccessLog = config.components['accesslog'][accesslog]

    # Load everything else.
    config.debug = rc.get('debug', default=False)
    config.config_path = rc.get('config_path', cast=pathlib.Path, exists=True)
    config.server_url = rc.get('server_url')
    config.scope_prefix = rc.get('scope_prefix')
    config.scope_formatter = rc.get('scope_formatter', cast=importstr)
    config.scope_max_length = rc.get('scope_max_length', cast=int)
    config.default_auth_client = rc.get('default_auth_client')
    config.http_basic_auth = rc.get('http_basic_auth', default=False, cast=asbool)
    config.token_validation_key = rc.get('token_validation_key', cast=json.loads) or None
    config.datasets = rc.get('datasets', default={})
    config.env = rc.get('env')
    config.docs_path = rc.get('docs_path', default=None)
    config.always_show_id = rc.get('always_show_id', default=False)
    config.root = rc.get('root', default=None)
    if config.root is not None:
        config.root = config.root.strip().strip('/')

    return config


@check.register(Context, components.Config)
def check(context: Context, config: components.Config):
    if config.default_auth_client and not (config.config_path / 'clients' / f'{config.default_auth_client}.yml').exists():
        clients_dir = config.config_path / 'clients'
        raise Exception(f"default_auth_client {config.default_auth_client!r} does not exist in {clients_dir}.")



from typing import List

import datetime

from sqlalchemy.engine.result import RowProxy

from spinta import commands
from spinta.components import Context, Model, Action
from spinta.types.datatype import Date, DateTime
from spinta.backends.postgresql.components import PostgreSQL


@commands.prepare_data_for_response.register(Context, Action, Model, PostgreSQL, RowProxy)
def prepare_data_for_response(
    context: Context,
    action: Action,
    model: Model,
    backend: PostgreSQL,
    value: RowProxy,
    *,
    select: List[str] = None,
) -> dict:
    return commands.prepare_data_for_response(
        context,
        action,
        model,
        backend,
        dict(value),
        select=select,
    )


@commands.prepare.register(Context, DateTime, PostgreSQL, datetime.datetime)
def prepare(
    context: Context,
    dtype: DateTime,
    backend: PostgreSQL,
    value: datetime.datetime,
) -> object:
    # convert datetime object to isoformat string if it belongs
    # to a nested property
    if dtype.prop.parent is dtype.prop.model:
        return value
    else:
        return value.isoformat()


@commands.prepare.register(Context, Date, PostgreSQL, datetime.date)
def prepare(context: Context, dtype: Date, backend: PostgreSQL, value: datetime.date) -> object:
    # convert date object to isoformat string if it belongs
    # to a nested property
    if dtype.prop.parent is dtype.prop.model:
        return value
    else:
        return value.isoformat()



from pathlib import Path
from typing import Optional

from typer import Argument

from spinta.manifests.tabular.helpers import write_tabular_manifest


def init(
    manifest: Optional[Path] = Argument(None),
):
    """Initialize a new manifest table"""
    with manifest.open('w') as f:
        write_tabular_manifest(f, [])




from typing import Any
from typing import Dict
from typing import Iterator
from typing import List
from typing import Optional
from typing import Tuple

import frictionless
import sqlalchemy as sa
from sqlalchemy.engine.base import Engine as SaEngine

from spinta import commands
from spinta import spyna
from spinta.backends import Backend
from spinta.components import Context
from spinta.datasets.backends.sql.components import Sql
from spinta.datasets.backends.sql.ufuncs.components import Engine
from spinta.datasets.backends.sql.ufuncs.components import SqlResource
from spinta.datasets.components import Resource
from spinta.exceptions import UnexpectedFormulaResult
from spinta.manifests.components import Manifest
from spinta.manifests.helpers import load_manifest_nodes
from spinta.utils.imports import full_class_name
from spinta.utils.naming import to_model_name
from spinta.utils.naming import to_property_name


def _ensure_list(value: Optional[Any]):
    if value is None:
        return value
    elif isinstance(value, list):
        return value
    else:
        return [value]


def _read_foreign_keys(
    model_name: str,
    resource: Optional[Resource],
    foreign_keys: List[Dict[str, Any]],
    properties: Dict[str, dict],
) -> Dict[str, Any]:
    for fk in foreign_keys:
        if len(fk['fields']) == 1:
            name = to_property_name(fk['fields'][0])
            prop = properties[name]
        else:
            name = to_property_name('_'.join(fk['fields']))
            prop = properties[name] = {}
            prop['prepare'] = ', '.join([
                to_property_name(f) for f in fk['fields']
            ])

        ref_model_name = to_model_name(fk['reference']['resource'])
        if ref_model_name:
            if resource:
                ref_model_name = f'{resource.dataset.name}/{ref_model_name}'
        else:
            # If ref_model_name is empty string, it means, this is a self
            # refrence.
            ref_model_name = model_name

        prop['type'] = 'ref'
        prop['model'] = ref_model_name
        prop['refprops'] = [
            to_property_name(f)
            for f in fk['reference']['fields']
        ]

    return properties


def _read_frictionless_field(field: frictionless.Field) -> Dict[str, Any]:
    return {
        'type': field.type,
        'external': {
            'name': field.name,
        }
    }


def _read_frictionless_resource(
    resource: Optional[Resource],
    backend: Backend,
    frictionless_resource: frictionless.Resource,
) -> Dict[str, Any]:
    schema = frictionless_resource.schema
    name = to_model_name(frictionless_resource.name)
    if resource:
        name = f'{resource.dataset.name}/{name}'
    return {
        'type': 'model',
        'name': name,
        'external': {
            'dataset': resource.dataset.name if resource else None,
            'resource': resource.name if resource else None,
            'name': frictionless_resource.name,
            'pk': [
                to_property_name(p)
                for p in _ensure_list(schema.primary_key)
            ],
        },
        'properties': _read_foreign_keys(
            name,
            resource,
            schema.foreign_keys,
            {
                to_property_name(field.name): _read_frictionless_field(field)
                for field in schema.fields
            },
        ),
    }


def _read_frictionless_package(
    resource: Optional[Resource],
    backend: Backend,
    package: frictionless.Package,
) -> Iterator[Tuple[int, Dict[str, Any]]]:
    for i, frictionless_resource in enumerate(package.resources):
        yield i, _read_frictionless_resource(
            resource,
            backend,
            frictionless_resource,
        )


@commands.inspect.register(Context, Manifest, Sql)
def inspect(context: Context, manifest: Manifest, backend: Sql):
    engine = sa.create_engine(backend.config['dsn'])
    package = frictionless.Package.from_sql(engine=engine)
    schemas = _read_frictionless_package(None, backend, package)
    load_manifest_nodes(context, manifest, schemas)


@commands.inspect.register(Context, Resource, Sql)
def inspect(context: Context, resource: Resource, backend: Sql):
    if resource.prepare:
        env = SqlResource(context).init(backend.config['dsn'])
        engine = env.resolve(resource.prepare)
        engine: Engine = env.execute(engine)
        if not isinstance(engine, Engine):
            raise UnexpectedFormulaResult(
                resource,
                formula=spyna.unparse(resource.prepare),
                expected=full_class_name(SaEngine),
                received=full_class_name(engine),
            )
    else:
        engine = Engine(backend.config['dsn'])

    package = frictionless.Package.from_sql(
        engine=engine.create(),
        namespace=engine.schema,
    )
    schemas = _read_frictionless_package(resource, backend, package)
    load_manifest_nodes(
        context,
        resource.dataset.manifest,
        schemas,
        link=True,
    )



from __future__ import annotations

from typing import Any, Dict, List, Optional, Tuple, Union

import collections
import logging
import os
import pathlib
import enum
import sys

from ruamel.yaml import YAML
import pkg_resources as pres

from spinta.utils.imports import importstr
from spinta.utils.schema import NA

Schema = Dict[str, Any]
Key = Tuple[str]

yaml = YAML(typ='safe')

log = logging.getLogger(__name__)

SCHEMA = {
    'type': 'object',
    'items': yaml.load(
        pathlib.Path(pres.resource_filename('spinta', 'config.yml')).
        read_text()
    ),
}


def read_config(args=None, envfile=None):
    rc = RawConfig()
    rc.read([
        Path('spinta', 'spinta.config:CONFIG'),
        EnvFile('envfile', envfile or '.env'),
        EnvVars('envvars', os.environ),
        CliArgs('cliargs', args or []),
    ])

    # Inject extension provided defaults
    configs = rc.get('config', cast=list, default=[])
    if configs:
        rc.read([Path(c, c) for c in configs], after='spinta')

    return rc


class KeyFormat(str, enum.Enum):
    cfg = 'cfg'
    cli = 'cli'
    env = 'env'


class ConfigSource:
    name: str

    def __init__(self, name=None, config=None):
        self.name = self.getname(name)
        self.config = config

    def __str__(self):
        return self.name

    def __repr__(self):
        return type(self).__module__ + '.' + type(self).__name__ + '(' + repr(self.name) + ')'

    def getname(self, name):
        return name or self.name or type(self).__name__

    def read(self, schema: Schema):
        config = {}
        for k, v in self.config.items():
            v = dict(_traverse(v, k))
            v.update(_get_inner_keys(v, depth=len(k)))
            config.update(v)
        self.config = config

    def keys(self, env: str = None):
        if env:
            for key in self.config:
                if key[:2] == ('environments', env):
                    yield key[2:]
        else:
            for key in self.config:
                if key[:1] != ('environments',):
                    yield key

    def get(self, key: tuple, env: str = None):
        if env:
            return self.config.get(('environments', env) + key, NA)
        else:
            return self.config.get(key, NA)


class PyDict(ConfigSource):

    def read(self, schema: Schema):
        envs = self.config.pop('environments', {})
        config = {
            tuple(k.split('.')): v
            for k, v in self.config.items()
        }
        for env, values in envs.items():
            for k, v in values.items():
                config[('environments', env) + tuple(k.split('.'))] = v
        self.config = config
        super().read(schema)


class Path(PyDict):

    def read(self, schema: Schema):
        if self.config.endswith(('.yml', '.yaml')):
            path = pathlib.Path(self.config)
            self.config = yaml.load(path.read_text())
        else:
            self.config = importstr(self.config)
        super().read(schema)


class CliArgs(PyDict):
    name = 'cli'

    def read(self, schema: Schema):
        config = {}
        for arg in self.config:
            key, val = arg.split('=', 1)
            if ',' in val:
                val = [v.strip() for v in val.split(',')]
            config[key] = val
        self.config = config
        super().read(schema)


class EnvVars(ConfigSource):
    name = 'env'

    def read(self, schema: Schema):
        config = {}
        for key, val in self.config.items():
            if not key.startswith('SPINTA_'):
                continue
            key = key[len('SPINTA_'):]
            key = tuple(key.lower().split('__'))
            if len(key) > 1 and key[0] not in schema['items'] and key[1] in schema['items']:
                key = ('environments',) + key
            config[key] = val
        self.config = config
        super().read(schema)


class EnvFile(EnvVars):

    def read(self, schema: Schema):
        config = {}
        path = pathlib.Path(self.config)
        if path.exists():
            with path.open() as f:
                for line in f:
                    if line.startswith('#'):
                        continue
                    line = line.strip()
                    if line == '':
                        continue
                    if '=' not in line:
                        continue
                    name, value = line.split('=', 1)
                    config[name] = value
        self.config = config
        super().read(schema)


class RawConfig:
    sources: List[ConfigSource]

    def __init__(self, sources=None):
        self._locked = False
        self.sources = sources or []
        self._keys: Dict[Tuple[str], Tuple[int, List[str]]] = {}
        self._schema = SCHEMA

    def read(
        self,
        sources: List[ConfigSource],
        after: Optional[str] = None,
    ):
        if self._locked:
            raise Exception(
                "Configuration is locked, use `rc.fork()` if you need to "
                "change configuration."
            )

        for config in sources:
            log.info(f"Reading config from {config.name}.")
            config.read(self._schema)

        if after is not None:
            pos = (i for i, s in enumerate(self.sources) if s.name == after)
            pos = next(pos, None)
            if pos is None:
                raise Exception(f"Given after value {after!r} does not exist.")
            pos += 1
            self.sources[pos:pos] = sources
        else:
            self.sources.extend(sources)

        self._keys = self._update_keys()

    def add(self, name, params):
        self.read([PyDict(name, params)])
        return self

    def fork(self, sources=None, after=None) -> RawConfig:
        rc = RawConfig(list(self.sources))
        if sources:
            if isinstance(sources, dict):
                rc.add('fork', sources)
            else:
                rc.read(sources, after)
        else:
            rc._keys = rc._update_keys()
        return rc

    def lock(self):
        self._locked = True

    def get(self, *key, default=NA, cast=None, required=False, exists=False, origin=False):
        env, _ = self._get_config_value(('env',), default=None)
        value, config = self._get_config_value(key, default, env)

        if cast is not None:
            if cast is list and isinstance(value, str):
                value = value.split(',') if value else []
            elif value is not None:
                value = cast(value)
            else:
                # XXX: why []?
                value = default or []

        if required and not value:
            name = '.'.join(key)
            raise Exception(f"{name!r} is a required configuration option.")

        if exists and isinstance(value, pathlib.Path) and not value.exists():
            name = '.'.join(key)
            raise Exception(f"{name} ({value}) path does not exist.")

        if origin:
            if config:
                return value, config.name
            else:
                return value, ''
        else:
            return value

    def keys(self, *key, origin=False) -> Union[
        List[str],
        Tuple[List[str], str]
    ]:
        config, keys = self._keys.get(key, (None, []))
        return (keys, config.name) if origin else keys

    def getall(self, *key, origin=False):
        keys = self.keys(*key)
        if keys:
            for k in keys:
                yield from self.getall(*key, k, origin=origin)
        else:
            res = self.get(*key, origin=origin)
            res = res if origin else (res,)
            yield (key,) + res

    def dump(self, *names, fmt: KeyFormat = KeyFormat.cfg, file=sys.stdout):
        table = [('Origin', 'Name', 'Value')]
        sizes = [len(x) for x in table[0]]
        for key, val, origin in self.getall(origin=True):
            if names:
                for name in names:
                    it = enumerate(name.split('.'))
                    if all(key[i].startswith(k) for i, k in it if k):
                        break
                else:
                    continue

            if fmt == KeyFormat.env:
                key = 'SPINTA_' + '__'.join(key).upper()
            else:
                key = '.'.join(key)

            if isinstance(val, list):
                for i, v in enumerate(val):
                    row = (origin, key + f'.{i}', v)
                    table.append(row)
                    sizes = [max(x) for x in zip(sizes, map(len, map(str, row)))]
            else:
                row = (origin, key, val)
                table.append(row)
                sizes = [max(x) for x in zip(sizes, map(len, map(str, row)))]

        table = (
            table[:1] +
            [tuple(['-' * s for s in sizes])] +
            table[1:]
        )
        if file:
            for row in table:
                print('  '.join([str(x).ljust(s) for x, s in zip(row, sizes)]), file=file)
        else:
            return table

    def to_dict(self, *names: str) -> Dict[str, Any]:
        result = {}
        for key, val in self.getall(*names):
            key = '.'.join(key[len(names):])
            result[key] = val
        return result

    def _update_keys(self) -> Dict[Key, List[str]]:
        """Update inner keys respecting already set values."""
        keys = {}
        env, _ = self._get_config_value(('env',), default=None)
        for config in self.sources:
            self._update_config_keys(keys, config, config.keys())
            if env:
                self._update_config_keys(keys, config, config.keys(env), env)
        return keys

    def _update_config_keys(self, keys, config, ckeys, env=None):
        # Update `keys` in place.
        if () not in keys:
            keys[()] = config, []
        for key in ckeys:
            if key and key[0] not in keys[()][1]:
                keys[()][1].append(key[0])
            n = len(key)
            schema = self._schema
            for i in range(1, n + 1):
                schema = self._get_key_schema(schema, key[i - 1])
                if schema is None or schema['type'] != 'object':
                    # Skip all non object keys, only objects can have keys.
                    break
                k = tuple(key[:i])
                v = config.get(k, env)
                if v is not NA:
                    # Source has explicit value set.
                    if isinstance(v, str):
                        v = [x.strip() for x in v.split(',')]
                    else:
                        v = list(v)
                    keys[k] = config, v
                elif i < n:
                    # No explicit value set, just collect all parents.
                    if k not in keys:
                        keys[k] = config, []
                    if key[i] not in keys[k][1]:
                        keys[k][1].append(key[i])

    def _get_key_schema(self, schema: Schema, key: str):
        if schema['type'] == 'object':
            if 'items' in schema:
                if key in schema['items']:
                    return schema['items'][key]
            if 'case' in schema:
                for items in schema['case'].values():
                    if key in items:
                        return items[key]
            if 'keys' in schema and schema['keys']['type'] == 'string':
                return schema['values']

    def _get_config_value(self, key: Key, default: Any = NA, env: str = None):
        assert isinstance(key, tuple)
        for config in reversed(self.sources):
            val = NA
            if env:
                val = config.get(key, env)
            if val is NA:
                val = config.get(key)
            if val is not NA:
                return val, config
        if default is NA:
            schema = self._schema
            for k in key:
                schema = self._get_key_schema(schema, k)
                if schema is None:
                    break
            else:
                default = schema.get('default', NA)
            if default is NA:
                default = None
        return default, None

    def get_source_names(self) -> List[str]:
        return [source.name for source in self.sources]


def _traverse(value, path=()):
    if isinstance(value, dict):
        for k, v in value.items():
            yield from _traverse(v, path + (k,))
    else:
        yield path, value


def _get_inner_keys(config: Dict[tuple, Any], depth=1):
    """Get inner keys for config.

    `config` is flattened dict, that looks like this:

        {
            ('a', 'b', 'c'): 1,
            ('a', 'b', 'd'): 2,
        }

    Nested version of this would look like this:

        {
            'a': {
                'b': {
                    'c': 1,
                    'd': 1,
                }
            }
        }

    Then, the purpose of this function is to add keys to all inner nesting
    levels. For this example, function result will be:

        {
            ('a'): ['b'],
            ('a', 'b'): ['c', 'd'],
        }

    This is needed for `RawConfig` class, in order to be able to do things like
    this:

        config.keys('a', 'b')
        ['c', 'd']

    And this functionality is needed, because of environment variables. For
    example, in order to add a new backend, first you need to add new keys, like
    this:

        SPINTA_A=b,x

    And then, you can add values to it:

        SPINTA_A_X=3

    And the end configuration will look like this:

        {
            'a': {
                'b': {
                    'c': 1,
                    'd': 1,
                },
                'x': '3'
            }
        }

    """
    inner = collections.defaultdict(list)
    for key in config.keys():
        for i in range(depth, len(key)):
            k = tuple(key[:i])
            if key[i] not in inner[k]:
                inner[k].append(key[i])
    return inner


def _get_from_prefix(config: dict, prefix: tuple):
    for k, v in config.items():
        if k[:len(prefix)] == prefix:
            yield k[len(prefix):], v



from pathlib import Path
from typing import List

import click

from spinta import commands
from spinta.components import Context
from spinta.components import Mode
from spinta.components import Store
from spinta.core.config import RawConfig


def load_store(context: Context) -> Store:
    config = context.get('config')
    commands.load(context, config)
    commands.check(context, config)
    store = context.get('store')
    commands.load(context, store)
    return store


def prepare_manifest(context: Context, *, verbose: bool = True) -> Store:
    store = load_store(context)
    if verbose:
        click.echo(f"Loading manifest {store.manifest.name}...")
    commands.load(context, store.manifest)
    commands.link(context, store.manifest)
    commands.check(context, store.manifest)
    commands.wait(context, store)
    commands.prepare(context, store.manifest)
    return store


def configure(
    context: Context,
    manifests: List[Path],
    *,
    mode: Mode = Mode.internal,
) -> Context:
    context = context.fork('cli')

    if manifests:
        config = {
            'backends.run': {'type': 'memory'},
            'keymaps.default': {
                'type': 'sqlalchemy',
                'dsn': 'sqlite:///keymaps.db',
            },
            'manifests': {
                'run': {
                    'type': 'backend',
                    'backend': 'run',
                    'keymap': 'default',
                    'mode': mode.value,
                    'sync': [],
                },
            },
            'manifest': 'run',
        }

        for i, manifest_path in enumerate(manifests):
            manifest_name = f'run{i}'
            config['manifests'][manifest_name] = {
                'type': 'tabular',
                'path': manifest_path,
                'backend': '',
            }
            config['manifests']['run']['sync'].append(manifest_name)

        # Add given manifest file to configuration
        rc: RawConfig = context.get('rc')
        context.set('rc', rc.fork(config))

    return context



import pathlib
from typing import Optional
from typing import TextIO
from typing import Tuple

from typer import Argument
from typer import Context as TyperContext
from typer import Option
from typer import echo

from spinta import commands
from spinta.cli.helpers.auth import require_auth
from spinta.cli.helpers.store import load_store
from spinta.cli.helpers.store import prepare_manifest
from spinta.components import Context
from spinta.core.config import RawConfig
from spinta.manifests.components import Manifest
from spinta.manifests.tabular.helpers import datasets_to_tabular
from spinta.manifests.tabular.helpers import load_ascii_tabular_manifest
from spinta.manifests.tabular.helpers import render_tabular_manifest
from spinta.manifests.tabular.helpers import write_tabular_manifest


def _save_manifest(manifest: Manifest, dest: TextIO):
    # TODO: Currently saving is hardcoded to tabular manifest type, but it
    #       should be possible to save or probably freeze to any manifest type.
    rows = datasets_to_tabular(manifest)
    write_tabular_manifest(dest, rows)


def inspect(
    ctx: TyperContext,
    manifest: Optional[pathlib.Path] = Argument(None, help="Path to manifest."),
    resource: Optional[Tuple[str, str]] = Option((None, None), '-r', '--resource', help=(
        "Resource type and source URI (-r sql sqlite:////tmp/db.sqlite)"
    )),
    formula: str = Option('', '-f', '--formula', help=(
        "Formula if needed, to prepare resource for reading"
    )),
    output: Optional[pathlib.Path] = Option(None, '-o', '--output', help=(
        "Output tabular manifest in a specified file"
    )),
    auth: Optional[str] = Option(None, '-a', '--auth', help=(
        "Authorize as a client"
    )),
):
    """Update manifest schema from an external data source"""
    context: Context = ctx.obj
    context = context.fork('inspect')

    if any(resource):
        config = {
            'backends.null': {
                'type': 'memory',
            },
            'manifests.inspect': {
                'type': 'tabular',
                'backend': 'null',
                'keymap': '',
                'mode': 'external',
                'path': None,
            },
            'manifest': 'inspect',
        }

        # Add given manifest file to configuration
        rc: RawConfig = context.get('rc')
        context.set('rc', rc.fork(config))
        store = load_store(context)
        resource_type, resource_source = resource
        load_ascii_tabular_manifest(context, store.manifest, f'''
        d | r               | type            | source            | prepare
        dataset             |                 |                   |
          | {resource_type} | {resource_type} | {resource_source} | {formula}
        ''', strip=True)
        commands.check(context, store.manifest)
        commands.prepare(context, store.manifest)

    elif manifest:
        config = {
            'backends': [],
            'manifest': 'inspect',
            'manifests.inspect': {
                'type': 'tabular',
                'backend': '',
                'keymap': '',
                'mode': 'internal',
                'path': manifest,
            },
        }

        # Add given manifest file to configuration
        rc: RawConfig = context.get('rc')
        context.set('rc', rc.fork(config))

        # Load manifest
        store = prepare_manifest(context)

    else:
        # Load manifest
        store = prepare_manifest(context)

    manifest = store.manifest

    with context:
        require_auth(context, auth)
        commands.inspect(context, manifest)

    if output:
        with pathlib.Path(output).open('w') as f:
            _save_manifest(manifest, f)
    else:
        echo(render_tabular_manifest(manifest))



import os
import pathlib
import tempfile
from typing import Any

import boto3
import moto
import pytest
import sqlalchemy_utils as su
from responses import RequestsMock

from spinta.core.config import RawConfig
from spinta.core.config import read_config
from spinta.manifests.components import Manifest
from spinta.testing.cli import SpintaCliRunner
from spinta.testing.client import create_test_client
from spinta.testing.context import ContextForTests
from spinta.testing.context import create_test_context
from spinta.testing.datasets import Sqlite
from spinta.testing.manifest import compare_manifest


@pytest.fixture(scope='session')
def rc():
    with tempfile.TemporaryDirectory() as tmpdir:
        rc = read_config()
        rc.add('pytest', {
            'env': 'test',
            'keymaps.default': {
                'type': 'sqlalchemy',
                'dsn': 'sqlite:////' + os.path.join(tmpdir, 'keymaps.db'),
            },
        })
        rc.lock()
        yield rc


@pytest.fixture()
def sqlite():
    with tempfile.TemporaryDirectory() as tmpdir:
        yield Sqlite('sqlite:///' + os.path.join(tmpdir, 'db.sqlite'))


@pytest.fixture(scope='session')
def postgresql(rc):
    dsn = rc.get('backends', 'default', 'dsn', required=True)
    if su.database_exists(dsn):
        yield dsn
    else:
        su.create_database(dsn)
        yield dsn
        su.drop_database(dsn)


@pytest.fixture(scope='session')
def mongo(rc):
    yield
    dsn = rc.get('backends', 'mongo', 'dsn', required=False)
    db = rc.get('backends', 'mongo', 'db', required=False)
    if dsn and db:
        import pymongo
        client = pymongo.MongoClient(dsn)
        client.drop_database(db)


@pytest.fixture(scope='session')
def s3(rc):
    with moto.mock_s3():
        yield
        bucket_name = rc.get('backends', 's3', 'bucket', required=False)
        if bucket_name:
            s3 = boto3.resource('s3')
            s3_client = boto3.client('s3')
            try:
                objs = s3_client.list_objects(Bucket=bucket_name)
                contents = objs.get('Contents')
                if contents:
                    obj_keys = {'Objects': [
                        {'Key': obj['Key'] for obj in contents}
                    ]}
                    bucket = s3.Bucket(bucket_name)
                    bucket.delete_objects(Delete=obj_keys)
                    bucket.delete()
            except s3_client.exceptions.NoSuchBucket:
                pass


@pytest.fixture(scope='session')
def backends(postgresql, mongo, s3):
    yield {
        'postgresql': postgresql,
        'mongo': mongo,
        's3': s3,
    }


@pytest.fixture(scope='session')
def _context(rc: RawConfig, postgresql, mongo, s3):
    context: ContextForTests = create_test_context(rc)
    context.load()
    yield context


@pytest.fixture
def context(_context, mocker, tmpdir, request):
    with _context.fork('test') as context:
        store = context.get('store')
        if 'fs' in store.backends:
            # XXX: There must be a better way to provide tmpdir to fs backend.
            mocker.patch.object(store.backends['fs'], 'path', pathlib.Path(tmpdir))

        # In-memory accesslog used with spinta.accesslog.python.
        context.set('accesslog.stream', [])

        yield context

        # At this point, transaction must be closed, if it is not, then something is
        # wrong. Find out why transaction was not property closed.
        assert context.has('transaction') is False

        # If context was not loaded, then it means, that database was not touched.
        # All database operations require fully loaded context.
        if context.loaded:
            # XXX: Maybe instead of deleting everythin, we could rollback
            #      transactions, once this kind of functionality will be
            #      available? This should be more efficient.
            context.wipe_all()


@pytest.fixture
def responses():
    with RequestsMock() as mock:
        yield mock


@pytest.fixture
def app(context):
    context.attach('client', create_test_client, context)
    return context.get('client')


@pytest.fixture
def cli():
    return SpintaCliRunner(mix_stderr=False)


def pytest_addoption(parser):
    # TODO: Switch back to backend pytest param.
    #       We want to tests multiple backends and one backend can use
    #       multiple models.
    parser.addoption(
        "--model",
        action="append",
        default=[],
        help="run tests only for particular model ['postgres', 'mongo', 'postgres/datasets']",
    )


def pytest_configure(config):
    # https://docs.pytest.org/en/latest/mark.html#registering-marks
    config.addinivalue_line(
        "markers", "models(*models): mark test to run multiple times with each model specified"
    )


def pytest_generate_tests(metafunc):
    # Get model markers from test, if markers are set - leave test as is
    models = metafunc.definition.get_closest_marker('models')
    if not models:
        return

    # If there are markers, get them, together with model CLI options
    models = set(models.args)
    model_cli_options = set(metafunc.config.getoption('model'))

    # If model CLI options are not empty
    # then get common markers from test and CLI options
    if model_cli_options:
        models = models.intersection(model_cli_options)

    # Parametrize our test with calculated models.
    # If we pass to CLI model option, which does not have a test marker,
    # then pytest will skip the test all together.
    metafunc.parametrize('model', models)


@pytest.hookimpl(tryfirst=True)
def pytest_assertrepr_compare(op: str, left: Any, right: Any):
    if op == '==' and isinstance(left, Manifest) and isinstance(right, str):
        left, right = compare_manifest(left, right)
        return [f'{left!r} {op} {right!r}']



import pathlib

from spinta import commands
from spinta.components import Context
from spinta.manifests.yaml.components import YamlManifest
from spinta.manifests.yaml.helpers import yaml_config_params


@commands.configure.register(Context, YamlManifest)
def configure(context: Context, manifest: YamlManifest):
    rc = context.get('rc')
    manifest.path = rc.get('manifests', manifest.name, 'path', cast=pathlib.Path, required=True)
    yaml_config_params(context, manifest)



from spinta import commands
from spinta.components import Context
from spinta.manifests.components import Manifest


@commands.inspect.register(Context, Manifest)
def inspect(context: Context, manifest: Manifest):
    for dataset in manifest.objects['dataset'].values():
        commands.inspect(context, dataset)



from spinta.core.config import RawConfig


def fix_s3_backend_issue(rc: RawConfig) -> RawConfig:
    # S3 backend is hardcoded globally and requires a fixture.
    # This rc override deletes S3 backend from config.
    return rc.fork({
        'backends': {
            'default': {
                'type': 'memory',
            }
        },
    })



from __future__ import annotations

import csv
import pathlib
import textwrap
from typing import Any
from typing import Callable
from typing import Dict
from typing import Iterable
from typing import Iterator
from typing import List
from typing import NamedTuple
from typing import Optional
from typing import Set
from typing import TextIO
from typing import Tuple
from typing import TypeVar
from typing import Union
from typing import cast

from spinta import commands
from spinta import spyna
from spinta.backends import Backend
from spinta.backends.components import BackendOrigin
from spinta.components import Context
from spinta.dimensions.enum.components import EnumItem
from spinta.components import Model
from spinta.components import Namespace
from spinta.components import Property
from spinta.core.enums import Access
from spinta.core.ufuncs import unparse
from spinta.datasets.components import Dataset
from spinta.dimensions.enum.components import Enums
from spinta.dimensions.prefix.components import UriPrefix
from spinta.exceptions import PropertyNotFound
from spinta.manifests.components import Manifest
from spinta.manifests.helpers import load_manifest_nodes
from spinta.manifests.tabular.components import ACCESS
from spinta.manifests.tabular.components import BackendRow
from spinta.manifests.tabular.components import BaseRow
from spinta.manifests.tabular.components import DESCRIPTION
from spinta.manifests.tabular.components import DatasetRow
from spinta.manifests.tabular.components import ID
from spinta.manifests.tabular.components import MANIFEST_COLUMNS
from spinta.manifests.tabular.components import ManifestColumn
from spinta.manifests.tabular.components import ManifestRow
from spinta.manifests.tabular.components import ManifestTableRow
from spinta.manifests.tabular.components import ModelRow
from spinta.manifests.tabular.components import PREPARE
from spinta.manifests.tabular.components import PROPERTY
from spinta.manifests.tabular.components import PrefixRow
from spinta.manifests.tabular.components import PropertyRow
from spinta.manifests.tabular.components import REF
from spinta.manifests.tabular.components import ResourceRow
from spinta.manifests.tabular.components import SOURCE
from spinta.manifests.tabular.components import TITLE
from spinta.manifests.tabular.constants import DATASET
from spinta.types.datatype import Ref
from spinta.utils.data import take
from spinta.utils.schema import NA

ParsedRow = Tuple[int, Dict[str, Any]]


MAIN_DIMENSIONS = [
    'dataset',
    'resource',
    'base',
    'model',
    'property',
]
EXTRA_DIMENSIONS = [
    '',
    'prefix',
    'enum',
    'param',
    'comment',
    'ns',
]


class TabularManifestError(Exception):
    pass


def _detect_header(
    path: Optional[pathlib.Path],
    line: int,  # Line number
    row: Iterable[str],
) -> List[str]:
    header = [h.strip().lower() for h in row]
    unknown_columns = set(header[:len(DATASET)]) - set(DATASET)
    if unknown_columns:
        unknown_columns = ', '.join(sorted(unknown_columns, key=header.index))
        raise TabularManifestError(
            f"{path}:{line}: Unknown columns: {unknown_columns}."
        )
    return header


def _detect_dimension(
    path: Optional[pathlib.Path],
    line: int,  # Line number
    row: Dict[str, str],
) -> Optional[str]:
    dimensions = [k for k in MAIN_DIMENSIONS if row[k]]

    if len(dimensions) == 1:
        return dimensions[0]

    if len(dimensions) > 1:
        dimensions = ', '.join(dimensions)
        raise TabularManifestError(
            f"{path}:{line}: In one row only single dimension can be used, "
            f"but found more than one: {dimensions}"
        )

    if row['type']:
        if row['type'] not in EXTRA_DIMENSIONS:
            raise TabularManifestError(
                f"{path}:{line}:type: Unknown additional dimension name "
                f"{row['type']}."
            )
        return row['type']

    return ''


class TabularReader:
    state: State
    path: pathlib.Path
    line: int
    type: str
    name: str
    data: ManifestRow               # Used when `appendable` is False
    rows: List[Dict[str, Any]]      # Used when `appendable` is True
    appendable: bool = False        # Tells if reader is appendable.

    def __init__(
        self,
        state: State,
        path: pathlib.Path,
        line: int,
    ):
        self.state = state
        self.path = path
        self.line = line
        self.data = {}
        self.rows = []

    def __str__(self):
        return f"<{type(self).__name__} name={self.name!r}>"

    def read(self, row: Dict[str, str]) -> None:
        raise NotImplementedError

    def append(self, row: Dict[str, str]) -> None:
        if any(row.values()):
            self.error(
                f"Updates are not supported in context of {self.type!r}."
            )

    def release(self, reader: TabularReader = None) -> bool:
        raise NotImplementedError

    def items(self) -> Iterator[ParsedRow]:
        if self.appendable:
            for data in self.rows:
                yield self.line, data
        else:
            yield self.line, self.data

    def enter(self) -> None:
        raise NotImplementedError

    def leave(self) -> None:
        raise NotImplementedError

    def error(self, message: str) -> None:
        raise TabularManifestError(f"{self.path}:{self.line}: {message}")


class ManifestReader(TabularReader):
    type: str = 'manifest'
    datasets: Set[str]
    namespaces: Set[str]
    data: ManifestTableRow

    def read(self, row: ManifestRow) -> None:
        self.name = str(self.path)
        self.data = {
            'type': 'manifest',
        }

    def release(self, reader: TabularReader = None) -> bool:
        return reader is None

    def enter(self) -> None:
        self.datasets = set()
        self.namespaces = set()
        self.state.manifest = self

    def leave(self) -> None:
        self.state.manifest = None


class DatasetReader(TabularReader):
    type: str = 'dataset'
    data: DatasetRow

    def read(self, row: Dict[str, str]) -> None:
        self.name = row['dataset']

        if row['dataset'] in self.state.manifest.datasets:
            self.error("Dataset already defined.")

        self.data = {
            'type': 'dataset',
            'id': row['id'],
            'name': row['dataset'],
            'level': row['level'],
            'access': row['access'],
            'title': row['title'],
            'description': row['description'],
            'resources': {},
        }

    def release(self, reader: TabularReader = None) -> bool:
        return reader is None or isinstance(reader, (
            ManifestReader,
            DatasetReader,
        ))

    def enter(self) -> None:
        self.state.dataset = self

    def leave(self) -> None:
        self.state.dataset = None


class ResourceReader(TabularReader):
    type: str = 'resource'
    data: Union[BackendRow, ResourceRow]

    def read(self, row: Dict[str, str]) -> None:
        self.name = row['resource']

        if self.state.dataset is None:
            self.read_backend(row)
        else:
            self.read_resource(row)

    def read_backend(self, row: Dict[str, str]) -> None:
        # Backends will be loaded using
        # `spinta.manifests.helpers._load_manifest_backends`.

        if 'backends' not in self.state.manifest.data:
            self.state.manifest.data['backends'] = {}
        backends = self.state.manifest.data['backends']

        if self.name in backends:
            self.error(
                f"Backend {self.name!r} with the same name already defined."
            )

        self.data = {
            'type': row['type'],
            'name': self.name,
            'dsn': row['source'],
            'title': row['title'],
            'description': row['description'],
        }

        backends[self.name] = self.data

    def read_resource(self, row: Dict[str, str]) -> None:
        dataset = self.state.dataset.data

        if self.name in dataset['resources']:
            self.error("Resource with the same name already defined in ")

        self.data = {
            'type': row['type'],
            'backend': row['ref'],
            'external': row['source'],
            'prepare': (
                spyna.parse(row['prepare'])
                if row['prepare'] else None
            ),
            'level': row['level'],
            'access': row['access'],
            'title': row['title'],
            'description': row['description'],
        }

        dataset['resources'][self.name] = self.data

    def release(self, reader: TabularReader = None) -> bool:
        return reader is None or isinstance(reader, (
            ManifestReader,
            DatasetReader,
            ResourceReader,
        ))

    def enter(self) -> None:
        self.state.resource = self

    def leave(self) -> None:
        self.state.resource = None


class BaseReader(TabularReader):
    type: str = 'base'
    data: BaseRow

    def read(self, row: Dict[str, str]) -> None:
        self.name = row['base']

        dataset = self.state.dataset.data if self.state.dataset else None
        self.data = {
            'model': get_relative_model_name(dataset, row['base']),
            'pk': row['ref'],
        }

    def release(self, reader: TabularReader = None) -> bool:
        return reader is None or isinstance(reader, (
            ManifestReader,
            DatasetReader,
            ResourceReader,
            BaseReader,
        ))

    def enter(self) -> None:
        self.state.base = self

    def leave(self) -> None:
        self.state.base = None


class ModelReader(TabularReader):
    type: str = 'model'
    data: ModelRow

    def read(self, row: Dict[str, str]) -> None:
        dataset = self.state.dataset
        resource = self.state.resource
        base = self.state.base
        name = get_relative_model_name(
            dataset.data if dataset else None,
            row['model'],
        )

        if self.state.rename_duplicates:
            dup = 1
            _name = name
            while _name in self.state.models:
                _name = f'{name}_{dup}'
                dup += 1
            name = _name
        elif name in self.state.models:
            self.error(f"Model {name!r} with the same name is already defined.")

        self.name = name

        self.data = {
            'type': 'model',
            'id': row['id'],
            'name': name,
            'base': base.name if base else None,
            'level': row['level'],
            'access': row['access'],
            'title': row['title'],
            'description': row['description'],
            'properties': {},
            'external': {
                'dataset': dataset.name if dataset else '',
                'resource': resource.name if dataset and resource else '',
                'pk': (
                    [x.strip() for x in row['ref'].split(',')]
                    if row['ref'] else []
                ),
                'name': row['source'],
                'prepare': (
                    spyna.parse(row['prepare'])
                    if row['prepare'] else None
                ),
            },
        }

        if resource and not dataset:
            self.data['backend'] = resource.name

    def release(self, reader: TabularReader = None) -> bool:
        return reader is None or isinstance(reader, (
            ManifestReader,
            DatasetReader,
            ResourceReader,
            BaseReader,
            ModelReader,
        ))

    def enter(self) -> None:
        self.state.model = self
        self.state.models.add(self.name)

    def leave(self) -> None:
        self.state.model = None


def _parse_property_ref(ref: str) -> Tuple[str, List[str]]:
    if '[' in ref:
        ref = ref.rstrip(']')
        ref_model, ref_props = ref.split('[', 1)
        ref_props = [p.strip() for p in ref_props.split(',')]
    else:
        ref_model = ref
        ref_props = []
    return ref_model, ref_props


class PropertyReader(TabularReader):
    type: str = 'property'
    data: PropertyRow
    enums: Set[str]

    def read(self, row: Dict[str, str]) -> None:
        self.name = row['property']

        if self.state.model is None:
            context = self.state.stack[-1]
            self.error(
                f"Property {self.name!r} must be defined in a model context. "
                f"Now it is defined in {context.name!r} {context.type} context."
            )

        if row['property'] in self.state.model.data['properties']:
            self.error(
                f"Property {self.name!r} with the same name is already "
                f"defined for this {self.state.model.name!r} model."
            )

        self.data = {
            'type': row['type'],
            'prepare': (
                spyna.parse(row['prepare'])
                if row['prepare'] else None
            ),
            'level': row['level'],
            'access': row['access'],
            'uri': row['uri'],
            'title': row['title'],
            'description': row['description'],
        }

        dataset = self.state.dataset.data if self.state.dataset else None

        if row['type'] == 'ref':
            ref_model, ref_props = _parse_property_ref(row['ref'])
            self.data['model'] = get_relative_model_name(dataset, ref_model)
            self.data['refprops'] = ref_props

        if dataset or row['source']:
            self.data['external'] = {
                'name': row['source'],
                'prepare': self.data.pop('prepare'),
            }

        self.state.model.data['properties'][row['property']] = self.data

    def release(self, reader: TabularReader = None) -> bool:
        return reader is None or isinstance(reader, (
            ManifestReader,
            DatasetReader,
            ResourceReader,
            BaseReader,
            ModelReader,
            PropertyReader,
        ))

    def enter(self) -> None:
        self.state.prop = self

    def leave(self) -> None:
        self.state.prop = None


class AppendReader(TabularReader):
    type: str = 'append'
    data: ManifestRow

    def read(self, row: ManifestRow) -> None:
        self.name = row[REF]
        self.data = row

    def release(self, reader: TabularReader = None) -> bool:
        return True

    def enter(self) -> None:
        pass

    def leave(self) -> None:
        self.state.stack[-1].append(self.data)


class PrefixReader(TabularReader):
    type: str = 'prefix'
    data: PrefixRow

    def read(self, row: Dict[str, str]) -> None:
        self.name = row['ref']

        node = (
            self.state.prop or
            self.state.model or
            self.state.base or
            self.state.resource or
            self.state.dataset or
            self.state.manifest
        )

        if 'prefixes' not in node.data:
            node.data['prefixes'] = {}

        prefixes = node.data['prefixes']

        if self.name in prefixes:
            self.error(
                f"Prefix {self.name!r} with the same name is already "
                f"defined for this {node.name!r} {node.type}."
            )

        self.data = {
            'id': row['id'],
            'eid': f'{self.path}:{self.line}',
            'type': self.type,
            'name': self.name,
            'uri': row['uri'],
            'title': row['title'],
            'description': row['description'],
        }

        prefixes[self.name] = self.data

    def append(self, row: Dict[str, str]) -> None:
        self.read(row)

    def release(self, reader: TabularReader = None) -> bool:
        return not isinstance(reader, AppendReader)

    def enter(self) -> None:
        pass

    def leave(self) -> None:
        pass


class NamespaceReader(TabularReader):
    type: str = 'ns'
    appendable: bool = True

    def read(self, row: Dict[str, str]) -> None:
        if not row['ref']:
            # `ref` is a required parameter.
            return

        self.name = row['ref']

        manifest = self.state.manifest

        if self.name in manifest.namespaces:
            self.error(
                f"Namespace {self.name!r} with the same name is already "
                f"defined."
            )

        manifest.namespaces.add(self.name)

        self.rows.append({
            'id': row['id'],
            'type': self.type,
            'name': self.name,
            'title': row['title'],
            'description': row['description'],
        })

    def append(self, row: Dict[str, str]) -> None:
        self.read(row)

    def release(self, reader: TabularReader = None) -> bool:
        return not isinstance(reader, AppendReader)

    def enter(self) -> None:
        pass

    def leave(self) -> None:
        pass


def _read_enum_row(name: str, row: ManifestRow) -> Dict[str, Any]:
    return {
        'name': name,
        'source': row[SOURCE],
        'prepare': (
            spyna.parse(row[PREPARE])
            if row[PREPARE] else NA
        ),
        'access': row[ACCESS],
        'title': row[TITLE],
        'description': row[DESCRIPTION],
    }


class EnumReader(TabularReader):
    type: str = 'enum'
    appendable: bool = True

    def read(self, row: ManifestRow) -> None:
        prop = self.state.prop

        if row[REF]:
            self.name = row[REF]
        else:
            self.name = ''

        if 'enums' not in prop.data:
            prop.data['enums'] = {}

        if self.name in prop.data['enums']:
            self.error(
                f"Enum {self.name!r} with the same name is already "
                f"defined."
            )

        source = row[SOURCE] or row[PREPARE]
        prop.data['enums'][self.name] = {
            source: _read_enum_row(self.name, row)
        }

    def append(self, row: ManifestRow) -> None:
        if not row[SOURCE] and not row[PREPARE]:
            # At least source or prepare must be defined.
            return

        enum = cast(EnumReader, self.state.stack[-1])
        prop = self.state.prop
        source = row[SOURCE] or row[PREPARE]

        if source in prop.data['enums'][enum.name]:
            self.error(
                f"Enum {self.name!r} item {source!r} with the same value is "
                f"already defined."
            )

        prop.data['enums'][enum.name][source] = _read_enum_row(enum.name, row)

    def release(self, reader: TabularReader = None) -> bool:
        return not isinstance(reader, AppendReader)

    def enter(self) -> None:
        pass

    def leave(self) -> None:
        pass


READERS = {
    # Main dimensions
    'dataset': DatasetReader,
    'resource': ResourceReader,
    'base': BaseReader,
    'model': ModelReader,
    'property': PropertyReader,

    # Extra dimensions
    '': AppendReader,
    'prefix': PrefixReader,
    'ns': NamespaceReader,
    'enum': EnumReader,
}


class State:
    stack: List[TabularReader]

    backends: Dict[str, Dict[str, str]] = None

    models: Set[str]

    manifest: ManifestReader = None
    dataset: DatasetReader = None
    resource: ResourceReader = None
    base: BaseReader = None
    model: ModelReader = None
    prop: PropertyReader = None

    rename_duplicates: bool = False

    def __init__(self):
        self.stack = []
        self.models = set()

    def release(self, reader: TabularReader = None) -> Iterator[ParsedRow]:
        for parent in list(reversed(self.stack)):
            if parent.release(reader):
                if isinstance(parent, (
                    ManifestReader,
                    NamespaceReader,
                    DatasetReader,
                    ModelReader,
                )):
                    yield from parent.items()
                self.stack.pop()
                parent.leave()
            else:
                break

        if reader:
            reader.enter()
            self.stack.append(reader)


def _read_tabular_manifest_rows(
    path: Optional[pathlib.Path],
    rows: Iterator[List[str]],
    *,
    rename_duplicates: bool = True,
) -> Iterator[ParsedRow]:
    header = next(rows, None)
    if header is None:
        # Looks like an empty file.
        return
    header = _detect_header(path, 1, header)

    defaults = {k: '' for k in MANIFEST_COLUMNS}

    state = State()
    state.rename_duplicates = rename_duplicates
    reader = ManifestReader(state, path, 1)
    reader.read({})
    yield from state.release(reader)

    for i, row in enumerate(rows, 2):
        row = dict(zip(header, row))
        row = {**defaults, **row}
        dimension = _detect_dimension(path, i, row)
        Reader = READERS[dimension]
        reader = Reader(state, path, i)
        reader.read(row)
        yield from state.release(reader)

    yield from state.release()


def read_tabular_manifest(
    path: pathlib.Path,
    *,
    rename_duplicates: bool = False,
) -> Iterator[ParsedRow]:
    with path.open() as f:
        csv_reader = csv.reader(f)
        yield from _read_tabular_manifest_rows(
            path,
            csv_reader,
            rename_duplicates=rename_duplicates,
        )


def striptable(table):
    return textwrap.dedent(table).strip()


def _join_escapes(row: List[str]) -> List[str]:
    res = []
    for v in row:
        if res and res[-1] and res[-1].endswith('\\'):
            res[-1] = res[-1][:-1] + '|' + v
        else:
            res.append(v)
    return res


def read_ascii_tabular_rows(
    manifest: str,
    *,
    strip: bool = False,
) -> Iterator[List[str]]:
    if strip:
        manifest = striptable(manifest)

    lines = (line.strip() for line in manifest.splitlines())
    lines = filter(None, lines)

    # Read header
    header = next(lines, None)
    if header is None:
        return
    header = normalizes_columns(header.split('|'))

    # Find index where dimension columns end.
    dim = sum(1 for h in header if h in DATASET[:6])
    yield header
    for line in lines:
        row = _join_escapes(line.split('|'))
        row = [x.strip() for x in row]
        rem = len(header) - len(row)
        row = row[:dim - rem] + [''] * rem + row[dim - rem:]
        assert len(header) == len(row), line
        yield row


def read_ascii_tabular_manifest(
    manifest: str,
    *,
    strip: bool = False,
    rename_duplicates: bool = False,
) -> Iterator[ParsedRow]:
    rows = read_ascii_tabular_rows(manifest, strip=strip)
    yield from _read_tabular_manifest_rows(
        None,
        rows,
        rename_duplicates=rename_duplicates,
    )


def load_ascii_tabular_manifest(
    context: Context,
    manifest: Manifest,
    manifest_ascii_table: str,
    *,
    strip: bool = False,
) -> None:
    schemas = read_ascii_tabular_manifest(manifest_ascii_table, strip=strip)
    load_manifest_nodes(context, manifest, schemas)
    commands.link(context, manifest)


def get_relative_model_name(dataset: dict, name: str) -> str:
    if name.startswith('/'):
        return name[1:]
    elif dataset is None:
        return name
    else:
        return '/'.join([
            dataset['name'],
            name,
        ])


def to_relative_model_name(model: Model, dataset: Dataset = None) -> str:
    """Convert absolute model `name` to relative."""
    if dataset is None:
        return model.name
    if model.name.startswith(dataset.name):
        prefix = dataset.name
        return model.name[len(prefix) + 1:]
    else:
        return '/' + model.name


def tabular_eid(model: Model):
    if isinstance(model.eid, int):
        return model.eid
    else:
        return 0


def _order_models_by_access(model: Model):
    return model.access or Access.private


class OrderBy(NamedTuple):
    func: Callable[[Union[Model, Property, EnumItem]], Any]
    reverse: bool = False


MODELS_ORDER_BY = {
    'access': OrderBy(_order_models_by_access, reverse=True),
    'default': OrderBy(tabular_eid),
}


def _order_properties_by_access(prop: Property):
    return prop.access or Access.private


PROPERTIES_ORDER_BY = {
    'access': OrderBy(_order_properties_by_access, reverse=True),
}


T = TypeVar('T', Model, Property, EnumItem)


def sort(
    ordering: Dict[str, OrderBy],
    items: Iterable[T],
    order_by: Optional[str],
) -> Iterable[T]:
    order: Optional[OrderBy] = None

    if order_by:
        order = ordering[order_by]
    elif 'default' in ordering:
        order = ordering['default']

    if order:
        return sorted(items, key=order.func, reverse=order.reverse)
    else:
        return items


def _prefixes_to_tabular(
    prefixes: Dict[str, UriPrefix],
) -> Iterator[ManifestRow]:
    first = True
    for name, prefix in prefixes.items():
        yield torow(DATASET, {
            'id': prefix.id,
            'type': prefix.type if first else '',
            'ref': name,
            'uri': prefix.uri,
            'title': prefix.title,
            'description': prefix.description,
        })
        first = False


def _backends_to_tabular(
    backends: Dict[str, Backend],
) -> Iterator[ManifestRow]:
    for name, backend in backends.items():
        yield torow(DATASET, {
            'type': backend.type,
            'resource': name,
            'source': backend.config.get('dsn'),
        })


def _namespaces_to_tabular(
    namespaces: Dict[str, Namespace],
) -> Iterator[ManifestRow]:
    namespaces = {
        k: ns
        for k, ns in namespaces.items() if not ns.generated
    }
    first = True
    for name, ns in namespaces.items():
        yield torow(DATASET, {
            'type': ns.type if first else '',
            'ref': name,
            'title': ns.title,
            'description': ns.description,
        })
        first = False


def _order_enums_by_access(item: EnumItem):
    return item.access or Access.private


ENUMS_ORDER_BY = {
    'access': OrderBy(_order_enums_by_access, reverse=True),
}


def _enums_to_tabular(
    enums: Optional[Enums],
    *,
    external: bool = True,
    access: Access = Access.private,
    order_by: ManifestColumn = None,
) -> Iterator[ManifestRow]:
    if enums is None:
        return
    for name, enum in enums.items():
        first = True
        items = sort(ENUMS_ORDER_BY, enum.values(), order_by)
        for item in items:
            if item.access < access:
                continue
            yield torow(DATASET, {
                'type': 'enum' if first else '',
                'ref': name if first else '',
                'source': item.source if external else '',
                'prepare': spyna.unparse(item.prepare or NA),
                'access': item.given.access,
                'title': item.title,
                'description': item.description,
            })
            first = False


def datasets_to_tabular(
    manifest: Manifest,
    *,
    external: bool = True,
    access: Access = Access.private,
    internal: bool = False,
    order_by: ManifestColumn = None,
) -> Iterator[ManifestRow]:
    yield from _prefixes_to_tabular(manifest.prefixes)
    yield from _backends_to_tabular(manifest.backends)
    yield from _namespaces_to_tabular(manifest.namespaces)
    dataset = None
    resource = None

    models = manifest.models if internal else take(manifest.models)
    models = sort(MODELS_ORDER_BY, models.values(), order_by)

    for model in models:
        if model.access < access:
            continue

        if model.external:
            if dataset is None or dataset.name != model.external.dataset.name:
                dataset = model.external.dataset
                if dataset:
                    resource = None
                    yield torow(DATASET, {
                        'id': dataset.id,
                        'dataset': dataset.name,
                        'level': dataset.level,
                        'access': dataset.given.access,
                        'title': dataset.title,
                        'description': dataset.description,
                    })

            if model.external and model.external.resource and (
                resource is None or
                resource.name != model.external.resource.name
            ):
                resource = model.external.resource
                if resource:
                    backend = resource.backend
                    yield torow(DATASET, {
                        'resource': resource.name,
                        'source': resource.external,
                        'prepare': unparse(resource.prepare or NA),
                        'type': resource.type,
                        'ref': (
                            backend.name
                            if (
                                backend and
                                backend.origin != BackendOrigin.resource
                            )
                            else ''
                        ),
                        'level': resource.level,
                        'access': resource.given.access,
                        'title': resource.title,
                        'description': resource.description,
                    })

        yield torow(DATASET, {})

        data = {
            'id': model.id,
            'model': model.name,
            'level': model.level,
            'access': model.given.access,
            'title': model.title,
            'description': model.description,
        }
        if model.external and model.external.dataset:
            data['model'] = to_relative_model_name(
                model,
                model.external.dataset,
            )
        if external and model.external:
            data.update({
                'source': model.external.name,
                'prepare': unparse(model.external.prepare or NA),
            })
            if (
                not model.external.unknown_primary_key and
                all(p.access >= access for p in model.external.pkeys)
            ):
                # Add `ref` only if all properties are available in the
                # resulting manifest.
                data['ref'] = ', '.join([
                    p.name for p in model.external.pkeys
                ])
        yield torow(DATASET, data)

        props = sort(PROPERTIES_ORDER_BY, model.properties.values(), order_by)
        for prop in props:
            if prop.name.startswith('_'):
                continue

            if prop.access < access:
                continue

            data = {
                'property': prop.place,
                'type': prop.dtype.name,
                'level': prop.level,
                'access': prop.given.access,
                'uri': prop.uri,
                'title': prop.title,
                'description': prop.description,
            }

            if external and prop.external:
                if isinstance(prop.external, list):
                    # data['source'] = ', '.join(x.name for x in prop.external)
                    # data['prepare'] = ', '.join(
                    #     unparse(x.prepare or NA)
                    #     for x in prop.external if x.prepare
                    # )
                    raise DeprecationWarning(
                        "Source can't be a list, use prepare instead."
                    )
                elif prop.external:
                    data['source'] = prop.external.name
                    data['prepare'] = unparse(prop.external.prepare or NA)

            if isinstance(prop.dtype, Ref):
                if model.external and model.external.dataset:
                    data['ref'] = to_relative_model_name(
                        prop.dtype.model,
                        model.external.dataset,
                    )
                    pkeys = prop.dtype.model.external.pkeys
                    rkeys = prop.dtype.refprops
                    if rkeys and pkeys != rkeys:
                        rkeys = ', '.join([p.place for p in rkeys])
                        data['ref'] += f'[{rkeys}]'
                else:
                    data['ref'] = prop.dtype.model.name

            yield torow(DATASET, data)
            yield from _enums_to_tabular(
                prop.enums,
                external=external,
                access=access,
                order_by=order_by,
            )


def torow(keys, values) -> ManifestRow:
    return {k: values.get(k) for k in keys}


def write_tabular_manifest(file: TextIO, rows: Iterable[ManifestRow]):
    writer = csv.DictWriter(file, fieldnames=DATASET)
    writer.writeheader()
    writer.writerows(rows)


def render_tabular_manifest(
    manifest: Manifest,
    cols: List[ManifestColumn] = None,
    *,
    sizes: Dict[ManifestColumn, int] = None,
) -> str:
    rows = datasets_to_tabular(manifest)
    return render_tabular_manifest_rows(rows, cols, sizes=sizes)


def render_tabular_manifest_rows(
    rows: Iterable[ManifestRow],
    cols: List[ManifestColumn] = None,
    *,
    sizes: Dict[ManifestColumn, int] = None,
) -> str:
    cols = cols or MANIFEST_COLUMNS
    hs = 1 if ID in cols else 0  # hierarchical cols start
    he = cols.index(PROPERTY)    # hierarchical cols end
    hsize = 1                      # hierarchical column size
    bsize = 3                      # border size
    if sizes is None:
        sizes = dict(
            [(c, len(c)) for c in cols[:hs]] +
            [(c, 1) for c in cols[hs:he]] +
            [(c, len(c)) for c in cols[he:]]
        )
        rows = list(rows)
        for row in rows:
            for i, col in enumerate(cols):
                val = '' if row[col] is None else str(row[col])
                if col == ID:
                    sizes[col] = 2
                elif i < he:
                    size = (hsize + bsize) * (he - hs - i) + sizes[PROPERTY]
                    if size < len(val):
                        sizes[PROPERTY] += len(val) - size
                elif sizes[col] < len(val):
                    sizes[col] = len(val)

    line = []
    for col in cols:
        size = sizes[col]
        line.append(col[:size].ljust(size))
    lines = [line]

    for row in rows:
        if ID in cols:
            line = [row[ID][:2] if row[ID] else '  ']
        else:
            line = []

        for i, col in enumerate(cols[hs:he + 1]):
            val = row[col] or ''
            if val:
                depth = i
                break
        else:
            val = ''
            depth = 0

        line += [' ' * hsize] * depth
        size = (hsize + bsize) * (he - hs - depth) + sizes[PROPERTY]
        line += [val.ljust(size)]

        for col in cols[he + 1:]:
            val = '' if row[col] is None else str(row[col])
            val = val.replace('|', '\\|')
            size = sizes[col]
            line.append(val.ljust(size))

        lines.append(line)

    lines = [' | '.join(line) for line in lines]
    lines = [l.rstrip() for l in lines]
    return '\n'.join(lines)


SHORT_NAMES = {
    'd': 'dataset',
    'r': 'resource',
    'b': 'base',
    'm': 'model',
    'p': 'property',
    't': 'type',
}


def normalizes_columns(cols: List[str]) -> List[ManifestColumn]:
    result: List[ManifestColumn] = []
    for col in cols:
        col = col.strip().lower()
        col = SHORT_NAMES.get(col, col)
        col = cast(ManifestColumn, col)
        if col in MANIFEST_COLUMNS:
            result.append(col)
        else:
            raise PropertyNotFound(property=col)
    return result



from typing import Any
from typing import Dict
from typing import List, Iterable, Optional
from typing import Tuple

import jsonpatch

from spinta import commands
from spinta.backends.components import BackendOrigin
from spinta.components import Mode
from spinta.backends.helpers import load_backend
from spinta.dimensions.prefix.helpers import load_prefixes
from spinta.nodes import get_node
from spinta.core.config import RawConfig
from spinta.components import Context, Config, Store, MetaData, EntryId
from spinta.manifests.components import Manifest
from spinta.manifests.internal.components import InternalManifest
from spinta.utils.enums import enum_by_name
from spinta.core.enums import Access


def create_manifest(
    context: Context,
    store: Store,
    name: str,
    seen: List[str] = None,
) -> Manifest:
    rc: RawConfig = context.get('rc')
    config = context.get('config')
    mtype = rc.get('manifests', name, 'type', required=True)
    Manifest = config.components['manifests'][mtype]
    manifest = Manifest()
    manifest.type = mtype
    _configure_manifest(context, rc, config, store, manifest, name, seen)
    return manifest


def create_internal_manifest(context: Context, store: Store) -> InternalManifest:
    rc = context.get('rc')
    config = context.get('config')
    manifest = InternalManifest()
    manifest.type = 'yaml'
    _configure_manifest(
        context, rc, config, store, manifest, 'internal',
        backend=store.manifest.backend.name if store.manifest.backend else None,
    )
    return manifest


def _configure_manifest(
    context: Context,
    rc: RawConfig,
    config: Config,
    store: Store,
    manifest: Manifest,
    name: str,
    seen: List[str] = None,
    *,
    backend: str = 'default',
):
    seen = seen or []
    manifest.name = name
    manifest.store = store
    manifest.parent = None
    manifest.access = rc.get('manifests', name, 'access') or 'protected'
    manifest.access = enum_by_name(manifest, 'access', Access, manifest.access)
    manifest.keymap = rc.get('manifests', name, 'keymap', default=None)
    manifest.keymap = store.keymaps[manifest.keymap] if manifest.keymap else None
    manifest.backend = rc.get('manifests', name, 'backend', default=backend)
    manifest.backend = store.backends[manifest.backend] if manifest.backend else None
    manifest.endpoints = {}
    manifest.backends = {}
    manifest.objects = {name: {} for name in config.components['nodes']}
    manifest.sync = []
    manifest.prefixes = {}
    manifest.mode = enum_by_name(manifest, 'mode', Mode, (
        rc.get('manifests', name, 'mode') or 'internal'
    ))
    for source in rc.get('manifests', name, 'sync', default=[], cast=list):
        if source in seen:
            raise Exception("Manifest sync cycle: " + ' -> '.join(seen + [source]))
        manifest.sync.append(
            create_manifest(context, store, source, seen + [source])
        )
    commands.configure(context, manifest)


def load_manifest_nodes(
    context: Context,
    manifest: Manifest,
    schemas: Iterable[Tuple[int, Optional[dict]]],
    *,
    source: Manifest = None,
    link: bool = False,
) -> None:
    to_link = []
    config = context.get('config')
    for eid, schema in schemas:
        if schema.get('type') == 'manifest':
            _load_manifest(context, manifest, schema, eid)
        else:
            node = _load_manifest_node(context, config, manifest, source, eid, schema)
            manifest.objects[node.type][node.name] = node
            if link:
                to_link.append(node)
    if to_link:
        for node in to_link:
            commands.link(context, node)


def _load_manifest_backends(
    context: Context,
    manifest: Manifest,
    backends: Dict[str, Dict[str, str]],
) -> None:
    manifest.backends = {}
    for name, data in backends.items():
        manifest.backends[name] = load_backend(
            context,
            manifest,
            name,
            BackendOrigin.manifest,
            data,
        )


def _load_manifest(
    context: Context,
    manifest: Manifest,
    data: Dict[str, Any],
    eid: EntryId,
) -> None:
    if 'backends' in data:
        _load_manifest_backends(context, manifest, data['backends'])
    if 'prefixes' in data:
        prefixes = load_prefixes(context, manifest, manifest, data['prefixes'])
        manifest.prefixes.update(prefixes)


def _load_manifest_node(
    context: Context,
    config: Config,
    manifest: Manifest,
    source: Optional[Manifest],
    eid: EntryId,
    data: dict,
) -> MetaData:
    node = get_node(config, manifest, eid, data)
    node.eid = eid
    node.type = data['type']
    node.parent = manifest
    node.manifest = manifest
    commands.load(context, node, data, manifest, source=source)
    return node


def get_current_schema_changes(
    context: Context,
    manifest: Manifest,
    eid: EntryId,
) -> List[dict]:
    freezed = commands.manifest_read_freezed(context, manifest, eid=eid)
    current = commands.manifest_read_current(context, manifest, eid=eid)
    patch = jsonpatch.make_patch(freezed, current)
    return list(patch)



from typing import List
from typing import Optional

from typer import Argument
from typer import Context as TyperContext
from typer import echo

from spinta.cli.helpers.store import prepare_manifest
from spinta.core.config import KeyFormat


def config(
    ctx: TyperContext,
    name: Optional[List[str]] = Argument(None),
    fmt: KeyFormat = KeyFormat.cfg,
):
    """Show current configuration values"""
    context = ctx.obj
    rc = context.get('rc')
    rc.dump(*name, fmt=fmt)


def check(
    ctx: TyperContext,
):
    """Check configuration and manifests"""
    context = ctx.obj
    prepare_manifest(context)
    echo("OK")



import csv
import pathlib
from typing import Union

from spinta import commands
from spinta.core.config import RawConfig
from spinta.manifests.components import Manifest
from spinta.manifests.helpers import load_manifest_nodes
from spinta.manifests.tabular.helpers import read_ascii_tabular_manifest
from spinta.manifests.tabular.helpers import read_ascii_tabular_rows
from spinta.manifests.tabular.helpers import read_tabular_manifest
from spinta.testing.context import create_test_context


def create_tabular_manifest(path: pathlib.Path, manifest: str):
    with path.open('w') as f:
        writer = csv.writer(f)
        rows = read_ascii_tabular_rows(manifest, strip=True)
        for row in rows:
            writer.writerow(row)


def load_tabular_manifest(
    rc: RawConfig,
    manifest: Union[pathlib.Path, str],
) -> Manifest:
    if isinstance(manifest, str) and '|' in manifest:
        path = ''
    else:
        path = manifest
        manifest = None

    rc = rc.fork({
        'manifest': 'default',
        'manifests': {
            'default': {
                'type': 'tabular',
                'path': str(path),
                'keymap': 'default',
            },
        },
    })

    context = create_test_context(rc)

    config = context.get('config')
    commands.load(context, config)

    store = context.get('store')
    commands.load(context, store)

    if path:
        schemas = read_tabular_manifest(path)
    else:
        schemas = read_ascii_tabular_manifest(manifest, strip=True)
    load_manifest_nodes(context, store.manifest, schemas)

    commands.link(context, store.manifest)

    return store.manifest



from dataclasses import dataclass

import sqlalchemy as sa
from sqlalchemy.engine.base import Engine as SaEngine

from spinta.core.ufuncs import Env
from spinta.utils.data import take
from spinta.utils.schema import NA


@dataclass
class Engine:
    dsn: str  # sqlalchemy engine dsn
    schema: str = None
    encoding: str = NA

    def create(self) -> SaEngine:
        return sa.create_engine(self.dsn, **take({
            'encoding': self.encoding,
        }))


class SqlResource(Env):
    dsn: str

    def init(self, dsn: str):
        return self(dsn=dsn)



from __future__ import annotations

from typing import List
from typing import Optional

from spinta.components import Property
from spinta.core.ufuncs import Expr
from spinta.types.datatype import DataType
from spinta.types.datatype import Ref


class ForeignProperty:
    """Representation of a reference.

    When querying `/city?select(country.code)`, `left` points to `country` and
    `right` points to `code`. `chain` will be:

        [
            ForeignProperty(country:Ref, code:String),
        ]

    Multiple references can be joined like this:

        /city?select(country.continent.planet.name)

    `chain` will look like this:

        [
            ForeignProperty(country:Ref, continent:Ref),
            ForeignProperty(continent:Ref, name:String),
        ]

    """

    left: Ref
    right: Optional[DataType]
    chain: List[ForeignProperty]
    name: str

    def __init__(
        self,
        fpr: Optional[ForeignProperty],
        left: Ref,
        right: DataType = None,
        *,
        # Used internally by swap method and only when fpr is None.
        _name: str = None,
        _chain: List[ForeignProperty] = None,
    ):
        if fpr is None:
            self.name = left.prop.place if _name is None else _name
            self.chain = [self] if _chain is None else (_chain + [self])
        else:
            self.name = fpr / left.prop
            self.chain = fpr.chain + [self]

        self.left = left
        self.right = right

    def __repr__(self):
        if self.right:
            place = self / self.right.prop
            return f'<{place}:{self.right.prop.name}>'
        else:
            return f'<{self.name}:None>'

    def __truediv__(self, prop: Property) -> str:
        return f'{self.name}->{prop.place}'

    def push(self, right: Optional[Property] = None):
        """Push another property to the chain

        Creates new ForeignProperty instance, by pushing another property to the
        chain. After this operation right will be pushed to left, and a new
        right will be placed instead of old one.
        """
        if isinstance(self.right, Ref):
            return ForeignProperty(self, self.right, right.dtype)
        elif self.right is None:
            return self.swap(right)
        else:
            raise RuntimeError(
                f"Can't push {right} to {self}, because right is a "
                f"{type(self.right)}, but it should be a Ref."
            )

    def swap(self, right: Optional[Property] = None):
        """Swap right with a new property

        Creates new ForeignProperty instance, preserving all properties, but
        change right to a new given right.
        """
        return ForeignProperty(
            None,
            self.left,
            right.dtype,
            _name=self.name,
            _chain=list(self.chain[:-1]),
        )

    def get_bind_expr(self) -> Expr:
        expr = self.right.get_bind_expr()
        for fpr in reversed(self.chain[:-1]):
            expr = Expr('getattr', fpr.right.get_bind_expr(), expr)
        return Expr('getattr', self.chain[0].left.get_bind_expr(), expr)



from spinta import commands
from spinta.components import Context
from spinta.datasets.components import Dataset


@commands.inspect.register(Context, Dataset)
def inspect(context: Context, dataset: Dataset):
    for resource in dataset.resources.values():
        commands.inspect(context, resource, resource.backend)



import datetime

from spinta import commands
from spinta.components import Context
from spinta.types.datatype import Date
from spinta.backends.mongo.components import Mongo


@commands.prepare.register(Context, Date, Mongo, datetime.date)
def prepare(context: Context, dtype: Date, backend: Mongo, value: datetime.date) -> datetime.datetime:
    # prepares date values for Mongo store, they must be converted to datetime
    return datetime.datetime.combine(value, datetime.datetime.min.time())



import itertools
import json

from starlette.requests import Request

from spinta import commands
from spinta.commands.formats import Format
from spinta.components import Context, Action, UrlParams, Node
from spinta import exceptions
from spinta.exceptions import NoBackendConfigured

METHOD_TO_ACTION = {
    'POST': Action.INSERT,
    'PUT': Action.UPDATE,
    'PATCH': Action.PATCH,
    'DELETE': Action.DELETE,
}


async def create_http_response(context: Context, params: UrlParams, request: Request):
    store = context.get('store')
    manifest = store.manifest

    if manifest.backend is None:
        raise NoBackendConfigured(manifest)

    if request.method == 'GET':
        context.attach('transaction', manifest.backend.transaction)

        if params.changes:
            _enforce_limit(context, params)
            return await commands.changes(context, request, params.model, params.model.backend, action=Action.CHANGES, params=params)

        elif params.pk:
            model = params.model
            action = Action.GETONE

            if params.prop and params.propref:
                prop = params.prop
                dtype = prop.dtype
                backend = model.backend
                return await commands.getone(context, request, prop, dtype, backend, action=action, params=params)
            elif params.prop:
                prop = params.prop
                dtype = prop.dtype
                backend = dtype.backend or model.backend
                return await commands.getone(context, request, prop, dtype, backend, action=action, params=params)
            else:
                backend = model.backend
                return await commands.getone(context, request, model, backend, action=action, params=params)

        else:
            search = any((
                params.query,
                params.select,
                params.limit is not None,
                params.offset is not None,
                params.sort,
                params.count,
            ))
            action = Action.SEARCH if search else Action.GETALL
            _enforce_limit(context, params)

            model = params.model
            backend = model.backend

            if backend is not None:
                # Namespace nodes do not have backend.
                context.attach(f'transaction.{backend.name}', backend.begin)

            if model.keymap:
                context.attach(f'keymap.{model.keymap.name}', lambda: model.keymap)

            return await commands.getall(context, request, model, backend, action=action, params=params)

    elif request.method == 'DELETE' and params.wipe:
        context.attach('transaction', manifest.backend.transaction, write=True)
        return await commands.wipe(context, request, params.model, params.model.backend, action=Action.WIPE, params=params)

    else:
        context.attach('transaction', manifest.backend.transaction, write=True)
        action = METHOD_TO_ACTION[request.method]
        if params.prop and params.propref:
            return await commands.push(context, request, params.prop.dtype, params.model.backend, action=action, params=params)
        elif params.prop:
            return await commands.push(context, request, params.prop.dtype, params.prop.dtype.backend, action=action, params=params)
        else:
            return await commands.push(context, request, params.model, params.model.backend, action=action, params=params)


def _enforce_limit(context: Context, params: UrlParams):
    config = context.get('config')
    fmt: Format = config.exporters[params.format]
    # XXX: I think this is not the best way to enforce limit, maybe simply
    #      an error should be raised?
    # XXX: Max resource count should be configurable.
    if not fmt.streamable and (params.limit is None or params.limit > 100):
        params.limit = 100
        params.limit_enforced = True


def peek_and_stream(stream):
    peek = list(itertools.islice(stream, 2))

    def _iter():
        for data in itertools.chain(peek, stream):
            yield data

    return _iter()


async def aiter(stream):
    for data in stream:
        yield data


async def get_request_data(node: Node, request: Request):
    ct = request.headers.get('content-type')
    if ct != 'application/json':
        raise exceptions.UnknownContentType(
            node,
            content_type=ct,
            supported_content_types=['application/json'],
        )

    try:
        data = await request.json()
    except json.decoder.JSONDecodeError as e:
        raise exceptions.JSONError(node, error=str(e))

    return data
